/*
	mapping.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2007,8 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"
#include "sys/time.h"

int main(int argc,char **argv)
{
	struct parameters* param = 0;
	char *time_s = 0;
	time_t start = 0;
	time_t end = 0;
	double time = 0;
	
	struct timeval tv1,tv2;
	struct timezone tz1,tz2;
	
	//fprintf(stderr,"%lud	long int\n%lud	long long\n%lud	int\n%lud	unsigned long int\n",sizeof(long int),sizeof(long long),sizeof(int),sizeof(unsigned long int));	
	
	param = interface(param,argc, argv);
	//if(param->time){
		gettimeofday(&tv1,&tz1);
		start = clock();
	//}
	
	if(param->input_clustering){
		get_mapped_expression(param);
		sort_db(param);
		if ((unlink(param->outfile_tmp)) == -1){
			perror("Removing temporary file failed");
			exit(-1);
		}
		
		cluster_mapping(param);
		if ((unlink(param->outfile_tmp)) == -1){
			perror("Removing temporary file failed");
			exit(-1);
		}
		
		if(!param->quiet_flag){
			time_s = get_time(time_s);
			fprintf(stderr,"%s	Exiting\n",time_s);
			end = clock();
			time = (double)(end-start)/CLOCKS_PER_SEC;
			gettimeofday(&tv2,&tz2);
			fprintf(stderr,"\n	Elapsed time:\n");
			fprintf(stderr,"	%7.0ld seconds (%0.2lf CPU+SYS seconds)\n",((tv2.tv_sec - tv1.tv_sec)  ),time);
		}
		free_param(param);
		exit(0);
	}
	
	if(param->write_suffix){
		create_suffix(param);
	}
	
	struct query_table* qt = 0 ;
	
	if (param->type_file){
		read_types_list(param);
	}
	
	if(param->outfile){
		if ((foutfile = fopen(param->outfile_tmp, "w")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
	}
	
	init_r();
	
	if(param->fasta_file && !param->fragment){
		set_hash_size(param);
	}
	
	if(!param->quiet_flag){
		time_s = get_time(time_s);
		fprintf(stderr,"\n%s	Reading in query sequences\n",time_s);
	}
	
	qt = init_query_hash(qt,param);
	
	if(param->fragment){
		qt = read_query_fasta_file_frag(qt,param,param->fragment);
	}else if(param->uniq){
		qt = read_query_fasta_file(qt,param);
	}else{
		qt = read_query_fasta_file_with_name(qt,param);
	}
	
	if(!param->quiet_flag){
		time_s = get_time(time_s);
		fprintf(stderr,"%s	Done\n",time_s);
	}
		
	//Need check for numseq here!!!!!!!
	if(!qt->size){
		fprintf(stderr,"No input sequences.. \n");
		free_query_hash(qt,param);
		free_param(param);
		exit(-1);
	}
	
	check_sequences(qt,param);
	
	if(param->solid){
		mapping_solid(qt, param);
	}else{
		mapping(qt,param);
	}
	
	if(param->stats){
		print_stats(qt,param);
		if(strcmp(param->stats,"STDOUT")){
			r_stats_pie(qt,param);
		}
	}
	
	if(param->ann){
		print_ann(qt,param);
		if(strcmp(param->ann,"STDOUT")){
			r_ann_pie(qt,param);
		}
		//print_ann_score(qt,param->max_annotation_overlap);
	}
	
	if(param->outfile){
		fclose(foutfile);
		if(param->fast || param->fragment){
			copy_mapping_output(qt,param);
		}else{
			if(param->uniq){
				clean_mapping_output(qt,param);
			}else{
				clean_name_mapping_output(qt,param);
			}
		}
		if((unlink(param->outfile_tmp)) == -1){
			perror("Removing temporary file failed");
			exit(-1);
		}
	}
	if(param->sort){
		sort_db(param);
		if ((unlink(param->outfile_tmp)) == -1){
			perror("Removing temporary file failed");
			exit(-1);
		}
	}
	
	if(!param->quiet_flag){
		time_s = get_time(time_s);
		fprintf(stderr,"%s	Exiting\n",time_s);
		report_stats(qt,param);
	}
	
	free_query_hash(qt,param);
	
	if(param->cluster != -1){
		cluster_mapping(param);
		if ((unlink(param->outfile_tmp)) == -1){
			perror("Removing temporary file failed");
			exit(-1);
		}
	}
	
	if(!param->quiet_flag){
		end = clock();
		time = (double)(end-start)/CLOCKS_PER_SEC;
		gettimeofday(&tv2,&tz2);
		fprintf(stderr,"	Elapsed time:\n");
		fprintf(stderr,"	%7.0ld seconds (%0.2lf CPU+SYS seconds)\n\n",((tv2.tv_sec - tv1.tv_sec)  ),time);
	}
	free_param(param);
#ifdef thread
	pthread_exit(NULL);
#endif
	return 0;
}

