/*
 *  mapping_solid.c
 *  NGSnexalign
 *
 *  Created by Timo Lassmann on 10/06/2008.
 *  Copyright 2008 RIKEN Yokohama Institute, Genome Exploration Research Group. All rights reserved.
 *
 */


#include "ngs_toolkit.h"
#include "sarray.h"

void mapping_solid(struct query_table* qt, struct parameters* param)
{
	struct fasta* target = 0;
	char *time_s = 0;
	
	FILE *file = 0;
	
	int db = 0;
	int i;
	
	clear_counts(qt);
	//target = alloc_target(target,param->solid_suffixfile,param->infiles);
	#ifdef thread
	struct thread_data* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct thread_data)*param->num_threads);
	assert(thread_data_array != 0);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	#endif
   	
	for(db = 0;db < param->infiles;db++){
		if(!param->quiet_flag){
			fprintf(stderr,"Searching against: %s\n",param->infile[db]);
			time_s = get_time(time_s);
		}
		
		if((file = fopen(param->solid_suffixfile[db],"r"))){
			fclose(file);
			if(!param->quiet_flag){
				fprintf(stderr,"%s	Found existing suffix file '%s'\n",time_s, param->solid_suffixfile[db]);
			}
			target = read_bin_aln(target,param->solid_suffixfile[db]);
			if(!target->mer_hash){
				target->mer_hash = malloc(sizeof(int)* 134217728);
				assert(target->mer_hash != 0);
				for(i = 0;i < 134217728;i++){
					target->mer_hash[i] = 0;
				}
				mer_hash(target,16);
			}
		}else{
			if(!param->quiet_flag){
				fprintf(stderr,"%s	Building SOLID suffix array...\n",time_s);
			}
			target = get_fasta(target,param->infile[db],param->solid);
			target->suffix = malloc((target->string_len+1)*sizeof(int));
			assert(target->suffix != 0);
			bsarray((unsigned char*)target->string, target->suffix, target->string_len);
			target->string = realloc(target->string,sizeof(char) * (target->string_len + 20));
			for (i = target->string_len+1;i < target->string_len+19;i++){
				target->string[i] = 'X';
			}
			target->string[i] = 0;
			target->string[target->string_len] = '!';
						
			target->mer_hash = malloc(sizeof(int)* 134217728);
			assert(target->mer_hash != 0);
			for(i = 0;i < 134217728;i++){
				target->mer_hash[i] = 0;
			}
			mer_hash(target,16);
		}	
		
		/*for(i = 12;i < 19;i++){
		 detect_mers(target, i);
		 }*/
		/*char * test = 0;
		fprintf(stdout,"%s\n",target->string);
		 for(i = 0; i < target->string_len;i++){
		 //fprintf(stdout,"%d	%d	%s	",i,target->suffix[i],target->string + target->suffix[i]);
		 test = target->string + target->suffix[i];
		 for(j = 0; j < 16;j++){
		 fprintf(stdout,"%c",test[j]);
		 }
		 fprintf(stdout,"\n");
		 }
		 exit(0);*/		
		
		for(i = 0; i < target->string_len;i++){
			if(target->suffix < 0){
				fprintf(stderr,"Damn = suffix is negative\n");
				exit(-1);
			}
		}
		//target->c = 0;
		if(!param->quiet_flag){
			time_s = get_time(time_s);
			fprintf(stderr,"%s	Done \n",time_s);
		}
		
		if(!target->suffix){
			fprintf(stderr,"No suffix\n");
			exit(-1);
		}
		
#ifdef thread
		rc = pthread_attr_init(&attr);
		if(rc){
			fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
			exit(-1);
		}
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
		interval =  (int)((double)qt->size/(double)param->num_threads);
		
		for(t = 0;t < param->num_threads - 1;t++) {
			thread_data_array[t].q = qt;
			thread_data_array[t].t = target;
			thread_data_array[t].start = t*interval;
			thread_data_array[t].end = t*interval + interval;
			thread_data_array[t].param = param;
		}
		thread_data_array[param->num_threads - 1].q = qt;
		thread_data_array[param->num_threads - 1].t = target;
		thread_data_array[param->num_threads - 1].start = (param->num_threads - 1)*interval;
		thread_data_array[param->num_threads - 1].end = qt->size;
		thread_data_array[param->num_threads - 1].param = param;
		
		
		if(param->count_matches || param->split_counts){
			for(t = 0;t < param->num_threads;t++) {
				rc = pthread_create(&threads[t], &attr, pattern_matching_solid_thread, (void *) &thread_data_array[t]);
				if (rc) {
					fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
					exit(-1);
				}
			}
		}else{
			for(t = 0;t < param->num_threads;t++) {
				rc = pthread_create(&threads[t], &attr, pattern_matching_solid_quick_thread, (void *) &thread_data_array[t]);
				if (rc) {
					fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
					exit(-1);
				}
			}
		}
		
		pthread_attr_destroy(&attr);
		
		for (t = 0;t < param->num_threads;t++){
			rc = pthread_join(threads[t], NULL);
			if (rc){
				fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
				exit(-1);
			}
		}
#else
		if(param->count_matches || param->split_counts){
			pattern_matching_solid(qt,target,param);
		}else{
			pattern_matching_solid_quick(qt,target,param);
		}
#endif
		if(param->split_counts){
			print_counts(qt,param,db);
			for(i= 0; i < qt->size;i++){
				qt->park[i]->stats[7] += qt->park[i]->stats[0];
			}
			//clear_counts(qt);
		}
		if(param->ann){
			set_class(qt,param,db);
		}
		if(param->ann || param->split_counts){
			clear_6_counts(qt);
		}
		free_aln(target);
	}
	//free_aln(target);
	
	if(param->count_matches){
		print_counts(qt,param,-1);
	}
	
	if(param->split_counts){
		for(i = 0;i < qt->size;i++){
			qt->park[i]->stats[0] = qt->park[i]->stats[7];
		}
	}
	
	if(param->ann){
		pop_ann_tree(qt,param->max_annotation_overlap);
		for(i = 0;i < qt->size;i++){
			qt->park[i]->stats[0] = qt->park[i]->stats[7];
		}		
	}
	
	if(param->unmapped_fasta){
		print_unmapped_fasta(qt,param->unmapped_fasta);
	}
	
	if(param->mapped_fasta){
		print_mapped_fasta(qt,param->mapped_fasta,param->upper_limit);
	}
	
	if(param->multi_mapped_fasta){
		print_multi_mapped_fasta(qt,param->multi_mapped_fasta,param->upper_limit);
	}
	
	//if(param->stats){
		pop_stats(qt,param);
	//}
	
#ifdef thread
	free(thread_data_array);
#endif
}
