/*
 *  patter_searching_short.c
 *  NGSnexalign
 *
 *  Created by Timo Lassmann on 10/06/2008.
 *  Copyright 2008 RIKEN Yokohama Institute, Genome Exploration Research Group. All rights reserved.
 *
 */

#include "ngs_toolkit.h"
#include "pattern_searching_short.h"
#include "mapping_output.h"


#ifdef thread

void* pattern_matching_short_thread(void *threadarg)
{
	struct thread_data *data;
	data = (struct thread_data *) threadarg;
	
	const struct query_table* qt = data->q;
	const struct fasta* t = data->t;
	
	const unsigned int upper_limit = data->param->upper_limit;
	const int start = data->start;
	const int end = data->end;
	const int format = data->param->format;
	const int mode = data->param->mode;
	struct query** tmp = qt->park;
	
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	int i;
	sp->mis = 0;
	sp->indel = 0;
	
	
	if(mode & 1){
		for(i = start; i < end;i++){
			sp = pm_short(sp,t,tmp[i]->seq,tmp[i]->len);
			tmp[i]->stats[0] += sp->sum;
			tmp[i]->stats[1] += sp->sum;
			//if(sp->sum > 0 && sp->sum <= upper_limit){
			if( tmp[i]->stats[0] <= upper_limit){
				mapping_output(sp,t,tmp[i],"E",format);
			}
		}
	}
	if(mode & 2){
		for(i = start; i < end;i++){
			pm_n_mis_short(t,tmp[i],tmp[i]->seq,upper_limit,format,0,data->param->mismatches,0,data->param->mismatches+1);
		}
	}
	if(mode & 4){
		for(i = start; i < end;i++){
			pm_ins_short(t,tmp[i],upper_limit,format);
			pm_del_short(t,tmp[i],upper_limit,format);
		}
	}
	
	free(sp);
	
	pthread_exit((void *) 0);
}

#else

void pattern_matching_short(struct query_table* qt,const struct fasta* t,const struct parameters* param)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp !=0);
	const unsigned int upper_limit = param->upper_limit;
	const int format = param->format;
	const int mode = param->mode;
	struct query** tmp = qt->park;
	
	int i,a;
	int len = 0;
	sp->mis = 0;
	sp->indel = 0;
	a = 1;
	if(mode & 1){
		for(i = 0; i < qt->size;i++){
			len = tmp[i]->len;
			sp = pm_short(sp,t,tmp[i]->seq,len);
			tmp[i]->stats[0] += sp->sum;
			tmp[i]->stats[1] += sp->sum;
			//if(sp->sum > 0 && sp->sum <= upper_limit){
			if(tmp[i]->stats[0] <= upper_limit){
				mapping_output(sp,t,tmp[i],"E",format);
			}
			//fprintf(stderr,"\r%8.0f percent done",(double)a /(double)qt->numseq * 100);
			a++;
		}
		//fprintf(stderr,"\n");
	}
	
	a = 1;
	if(mode & 2){
		for(i = 0; i < qt->size;i++){
			pm_n_mis_short(t,tmp[i],tmp[i]->seq,upper_limit,format,0,param->mismatches,0,param->mismatches+1);
		//	fprintf(stderr,"\r%8.0f percent done",(double)a /(double)qt->numseq * 100);
			a++;
		}
		//fprintf(stderr,"\n");
	}
	
	a = 1;
	if(mode & 4){
		for(i = 0; i < qt->size;i++){
			
			pm_ins_short(t,tmp[i],upper_limit,format);
			pm_del_short(t,tmp[i],upper_limit,format);
		//	fprintf(stderr,"\r%8.0f percent done",(double)a /(double)qt->numseq * 100);
			a++;
		}
		//fprintf(stderr,"\n");
	}
	
	free(sp);
}

#endif


struct suffix_pos* pm_short(struct suffix_pos* sp,const struct fasta*t ,const char *p,const int len)
{
	char rc[MAX_LINE];
	reverse_complement(p,rc,len);
	
	sp->e_p = binsearch_up(p,t->string,t->suffix,len,t->string_len);
	sp->s_p = binsearch_down(p,t->string,t->suffix,len,sp->e_p);
	
	sp->e_m = binsearch_up(rc,t->string,t->suffix,len,t->string_len);
	sp->s_m = binsearch_down(rc,t->string,t->suffix,len,sp->e_m);
	sp->sum = (sp->e_p-sp->s_p)+(sp->e_m-sp->s_m);
	return sp;
}

void pm_del_short(const struct fasta*t,struct query* q,const unsigned int upper_limit,const int format)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	char* tmp_seq = 0;
	char* mod_line = 0;
	int i,j,g;
	sp->mis = 0;
	sp->indel = 1;
	tmp_seq = malloc(sizeof(char)*(q->len));
	assert(tmp_seq != 0);
	mod_line = malloc(sizeof(char)*(20));
	assert(mod_line != 0);
	for(i = 0;i < q->len-1;i++){
		if(q->seq[i] != q->seq[i+1]){
			g = 0;// used to be 0
			for(j = 0;j < q->len;j++){//used to be 0
				if(j != i){
					tmp_seq[g] = q->seq[j];
					g++;
				}
			}
			tmp_seq[g] = 0;
			sp = pm_short(sp,t,tmp_seq,q->len-1);
			q->stats[0] += sp->sum;
			q->stats[3] += sp->sum;
			//if(sp->sum > 0 && sp->sum <= upper_limit){
			if(q->stats[0] <= upper_limit){
				sprintf(mod_line, "D%d%c",i,q->seq[i]);
				mapping_output(sp,t,q,mod_line,format);
			}
			
		}
	}
	g = 0;
	for(j = 0;j < q->len-1;j++){
		tmp_seq[g] = q->seq[j];
		g++;
	}
	tmp_seq[g] = 0;
	sp = pm_short(sp,t,tmp_seq,q->len-1);
	q->stats[0] += sp->sum;
	q->stats[3] += sp->sum;
	//if(sp->sum > 0 && sp->sum <= upper_limit){
	if(q->stats[0] <= upper_limit){
		sprintf(mod_line, "D%d%c",i,q->seq[i]);
		mapping_output(sp,t,q,mod_line,format);
	}
	
	
	free(tmp_seq);
	free(mod_line);
	free(sp);
}

void pm_ins_short(const struct fasta*t ,struct query* q,const unsigned int upper_limit,const int format)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp);
	char* alphabet = "ACGT";
	char* tmp_seq = 0;
	char* mod_line = 0;
	int i,j,f,g;
	sp->mis = 0;
	sp->indel = 1;
	tmp_seq = malloc(sizeof(char)*(q->len+2));
	assert(tmp_seq != 0);
	mod_line = malloc(sizeof(char)*(20));
	assert(mod_line != 0);
	for(i = 0;i < (q->len-2);i++){
		for(j = 0;j < 4;j++){
			if(q->seq[i+1] != alphabet[j]){
				g = 0;
				for(f = 0; f <= i;f++){
					tmp_seq[g] = q->seq[f];
					g++;
				}
				tmp_seq[g] = alphabet[j];
				g++;
				
				for(f = i+1;f < q->len;f++){
					tmp_seq[g] = q->seq[f];
					g++;
				}
				tmp_seq[g] = 0;
				sp = pm_short(sp,t,tmp_seq,q->len+1);
				q->stats[0] += sp->sum;
				q->stats[3] += sp->sum;
				//if(sp->sum > 0 && sp->sum <= upper_limit){
				if( q->stats[0] <= upper_limit){
					sprintf(mod_line, "I%d%c",i,alphabet[j]);
					mapping_output(sp,t,q,mod_line,format);
				}
				
			}
		}
	}
	i = q->len-2;
	for(j = 0;j < 4;j++){
		g = 0;
		for(f = 0; f <= i;f++){
			tmp_seq[g] = q->seq[f];
			g++;
		}
		tmp_seq[g] = alphabet[j];
		g++;
		
		for(f = i+1;f < q->len;f++){
			tmp_seq[g] = q->seq[f];
			g++;
		}
		tmp_seq[g] = 0;
		sp = pm_short(sp,t,tmp_seq,q->len+1);
		q->stats[0] += sp->sum;
		q->stats[3] += sp->sum;
		//if(sp->sum > 0 && sp->sum <= upper_limit){
		if( q->stats[0] <= upper_limit){
			sprintf(mod_line, "I%d%c",i,alphabet[j]);
			mapping_output(sp,t,q,mod_line,format);
		}
		
	}
	free(tmp_seq);
	free(mod_line);
	free(sp);
}

void pm_n_mis_short(const struct fasta*t ,struct query* q,const char *p,const unsigned int upper_limit,const int format,int start,int n,char* mod_line, int mis)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	char* alphabet = "ACGT";
	char* tmp_seq = 0;
	int m_col[5] = {-1,-1,2,4,5};
	int i,j,f,g;
	
	int mod_len = 0;
	sp->mis = 1;
	sp->indel = 0;
	tmp_seq = malloc(sizeof(char)*(q->len+1));
	assert(tmp_seq != 0);
	if(!mod_line){
		sp->indel = 1;
		mod_line = malloc(sizeof(char)*(10*n));
		assert(mod_line != 0);
	}else{
		mod_len = strlen(mod_line);
		mod_line[mod_len] = ',';
		mod_len++;
	}
	
	n--;
	for(i = start;i < q->len;i++){
		g = 0;
		for(f =0; f < i;f++){
			tmp_seq[g] = p[f];
			g++;
		}
		for(j = 0;j < 4;j++){
			if(alphabet[j] != p[i]){
				g = i;
				tmp_seq[g] = alphabet[j];
				g++;
				for(f = i+1;f < q->len;f++){
					tmp_seq[g] = p[f];
					g++;
				}
				tmp_seq[g] = 0;
					
				sprintf(mod_line+mod_len, "M%d%c%c",i,p[i],alphabet[j]);
				sp = pm_short(sp,t,tmp_seq,q->len);
				q->stats[m_col[mis-n]] += sp->sum;
				q->stats[0] += sp->sum;
				//if(sp->sum > 0 && sp->sum <= upper_limit){
				if( q->stats[0] <= upper_limit){
					sp->mis = mis-n-1;
					mapping_output(sp,t,q,mod_line,format);
				}
				
				if(n){
					pm_n_mis_short(t,q,tmp_seq,upper_limit,format,i+1,n,mod_line,mis);
				}
			}
		}
	}
	if(sp->indel){
		free(mod_line);
	}
	free(tmp_seq);
	free(sp);
}




